from .base_parser import BaseParser
import json
from scrapegraphai.graphs import SmartScraperGraph
import re
import aiohttp
from bs4 import BeautifulSoup
from youtube_transcript_api.formatters import TextFormatter
from src.content_parser.helpers.youtube_transcript_helper import YouTubeTranscriptHelper
from gitingest import ingest
from ..utils import ResultFormat
from typing import Any

class TextParser(BaseParser):
    @property
    def supported_mimetypes(self) -> list:
        return [
            "text/plain"
        ]

    def parse(self, source: Any, result_format: ResultFormat = ResultFormat.TXT) -> str:
        """Parse content and return as string.
        
        Args:
            file_content: Content to parse (bytes or str)
            result_format: Desired output format
            
        Returns:
            str: Parsed content as string
            
        Raises:
            ValueError: If content type is not supported
        """
        if isinstance(source, bytes):
            return source.decode('utf-8')
        elif isinstance(source, str):
            if source.startswith("http"):
                if "youtube.com" in source:
                    return self.transcribe_youtube(source, result_format)
                elif "github.com" in source:
                    return self.scrape_github(source, result_format)
                return self.scrape_url(source, result_format)
            return source
        raise ValueError("Unsupported content type")
    
    def scrape_url(self, url: str, result_format: ResultFormat) -> str:
        """Scrape URL content and return as string.
        
        Args:
            url: URL to scrape
            result_format: Desired output format
            
        Returns:
            str: Scraped content as JSON string
        """
        graph_config = {
            "llm": {
                "api_key": "",
                "model": "openai/gpt-4o-mini",
            },
            "verbose": True,
            "headless": True,
        }

        smart_scraper_graph = SmartScraperGraph(
            prompt="Give me all information from the website.",
            source=url,
            config=graph_config
        )

        result = smart_scraper_graph.run()
        return json.dumps(result, indent=4)
    
    def transcribe_youtube(self, url: str, result_format: ResultFormat) -> str:
        """Transcribe YouTube video and return as string.
        
        Args:
            url: YouTube video URL
            result_format: Desired output format
            
        Returns:
            str: Video title and transcript
            
        Raises:
            RegexResultError: If video ID cannot be extracted
            RemoteRequestError: If transcript/title cannot be fetched
        """
        helper = YouTubeTranscriptHelper()
        video_id = helper.extract_video_id(url)
        transcript = helper.get_best_transcript(video_id)
        title = helper.get_video_title(video_id)
        
        formatter = TextFormatter()
        formatted_transcript = formatter.format_transcript(transcript)
        
        return f"{title}\n\n{formatted_transcript}"

    def scrape_github(self, url: str, result_format: ResultFormat) -> str:
        """Scrape GitHub repository content.
        
        Args:
            url: GitHub repository URL
            result_format: Desired output format
            
        Returns:
            str: Repository content
        """
        summary, tree, content = ingest(url)
        return content



import base64
from enum import Enum
from urllib.parse import urlparse
import os


class ResultFormat(Enum):
    MD = "markdown"
    TXT = "txt"


def _is_url(source) -> bool:
    if not isinstance(source, str):
        return False
    parsed = urlparse(source)
    return parsed.scheme in ["http", "https"]

def _is_file_path(source) -> bool:
    return isinstance(source, str) and os.path.exists(source)

def _is_stream(source) -> bool:
    return hasattr(source, "read")

def _is_base64(source) -> bool:
    if not isinstance(source, str):
        return False
    try:
        base64.b64decode(source, validate=True)
        return True
    except Exception:
        return False



from loguru import logger

class ParseThisError(Exception):
    """Base class for exceptions in this module."""
    def __init__(self, message):
        logger.error(message)
        logger.exception(message)
        super().__init__(message)

class ParserNotFoundError(ParseThisError):
    """Exception raised when the magic parser detection wasnt able to find a parser."""
    pass

class UnsupportedMimeTypeError(ParseThisError):
    """Exception raised when the mime type is not supported by any parser."""
    pass

class RegexResultError(ParseThisError):
    """Exception raised when the regex search result is None."""
    pass

class RemoteRequestError(ParseThisError):
    """Exception raised for failing requests during parsing."""
    pass

class NotFoundError(ParseThisError):
    """Exception raised when the requested resource was not found."""
    pass

class ParsingFailed(ParseThisError):
    """Exception raised when parsing fails."""
    pass


import pytest
from unittest.mock import patch, MagicMock
from src.content_parser.text_parser import TextParser
import json
from src.parsethisio import ResultFormat
from src.exceptions import RegexResultError, RemoteRequestError, NotFoundError

mocked_scraper_response = '''{"title": "Example Domain", "description": "This domain is for use in illustrative examples in documents.", "usage": "You may use this domain in literature without prior coordination or asking for permission.", "more_information_link": "https://www.iana.org/domains/example"}'''

@pytest.fixture
def mock_scrapegraph_ai():
    with patch('src.content_parser.text_parser.SmartScraperGraph') as MockSmartScraperGraph:
        mock_client = MockSmartScraperGraph.return_value
        mock_client.run.return_value = json.loads(mocked_scraper_response)
        yield mock_client

def test_text_parser(mock_scrapegraph_ai):
    parser = TextParser()

    text = parser.parse("http://jdde.de/", result_format=ResultFormat.TXT)
    
    parsedJsonText = json.loads(text)

    #dump and load again to have the same attribute order
    adjustedResponse = mocked_scraper_response
    adjustedText = json.dumps(parsedJsonText)

    assert adjustedText == adjustedResponse
    mock_scrapegraph_ai.run.assert_called_once()

mocked_youtube_transcribe_title = "Tagesschau"
mocked_youtube_transcribe_answer = [{'text': 'This is the First German Television\nwith the Tagesschau.', 'start': 4.28, 'duration': 3.68}, {'text': 'This broadcast was\nsubtitled live by NDR (22.12.2024)', 'start': 8.68, 'duration': 4.44}, {'text': 'Today in the studio:\nConstantin Schreiber.', 'start': 15.52, 'duration': 2.64}, {'text': 'Good evening to the Tagesschau.', 'start': 18.36, 'duration': 2.08}]
exptected_transcribe_as_string = '''
This is the First German Television
with the Tagesschau.
This broadcast was
subtitled live by NDR (22.12.2024)
Today in the studio:
Constantin Schreiber.
Good evening to the Tagesschau.'''
@pytest.fixture
def mock_youtube_transcript_logic():
    with patch('src.content_parser.text_parser.YouTubeTranscriptHelper') as MockYouTubeTranscriptHelper:
        mock_client = MockYouTubeTranscriptHelper.return_value
        mock_client.get_best_transcript.return_value = mocked_youtube_transcribe_answer
        mock_client.get_video_title.return_value = mocked_youtube_transcribe_title
        yield mock_client

def test_text_parser_youtube_url(mock_youtube_transcript_logic):
    parser = TextParser()

    text = parser.parse("https://www.youtube.com/watch?v=Cc_Su-YC_h4", result_format=ResultFormat.TXT)

    expectedOutcome = mocked_youtube_transcribe_title + "\n" + exptected_transcribe_as_string

    assert text == expectedOutcome
    mock_youtube_transcript_logic.get_best_transcript.assert_called_once()
    mock_youtube_transcript_logic.get_video_title.assert_called_once()



mocked_github_scrape_ingest_content = '''
================================================
File: /CODE_OF_CONDUCT.md
================================================
# Contributor Covenant Code of Conduct

## Our Pledge

We as members, contributors, and leaders pledge to make participation in our
community a harassment-free experience for everyone, regardless of age, body
size, visible or invisible disability, ethnicity, sex characteristics, gender
identity and expression, level of experience, education, socio-economic status,
nationality, personal appearance, race, religion, or sexual identity
and orientation.
'''
@pytest.fixture
def mock_gitingest_answer():
    with patch('src.content_parser.text_parser.ingest') as MockIngest:
        MockIngest.return_value = ('', '', mocked_github_scrape_ingest_content)

        yield MockIngest

def test_text_parser_github_url(mock_gitingest_answer):
    parser = TextParser()

    text = parser.parse("https://github.com/cyclotruc/gitingest", result_format=ResultFormat.TXT)

    assert text == mocked_github_scrape_ingest_content
    mock_gitingest_answer.assert_called_once()

def test_text_parser_invalid_input():
    """Test that parser raises ValueError for invalid input types."""
    parser = TextParser()
    
    with pytest.raises(ValueError, match="Unsupported content type"):
        parser.parse(123)  # Test non-bytes, non-string input
    
    with pytest.raises(ValueError, match="Unsupported content type"):
        parser.parse(["not", "valid"])  # Test list input
    
    with pytest.raises(ValueError, match="Unsupported content type"):
        parser.parse({"key": "value"})  # Test dict input

@pytest.fixture
def mock_scrapegraph_ai_error():
    """Mock SmartScraperGraph to simulate network errors."""
    with patch('src.content_parser.text_parser.SmartScraperGraph') as MockSmartScraperGraph:
        mock_client = MockSmartScraperGraph.return_value
        mock_client.run.side_effect = Exception("Network error")
        yield mock_client

def test_text_parser_scrape_url_error(mock_scrapegraph_ai_error):
    """Test that parser handles network errors during URL scraping."""
    parser = TextParser()
    with pytest.raises(Exception, match="Network error"):
        parser.parse("http://example.com", result_format=ResultFormat.TXT)

def test_text_parser_youtube_invalid_url(mock_youtube_transcript_logic):
    """Test that parser handles invalid YouTube URLs."""
    parser = TextParser()
    mock_youtube_transcript_logic.extract_video_id.side_effect = RegexResultError("Invalid URL")
    
    with pytest.raises(RegexResultError, match="Invalid URL"):
        parser.parse("https://youtube.com/invalid", result_format=ResultFormat.TXT)

def test_text_parser_youtube_no_transcript(mock_youtube_transcript_logic):
    """Test that parser handles missing YouTube transcripts."""
    parser = TextParser()
    mock_youtube_transcript_logic.extract_video_id.return_value = "valid_id"
    mock_youtube_transcript_logic.get_best_transcript.side_effect = NotFoundError("No transcript")
    
    with pytest.raises(NotFoundError, match="No transcript"):
        parser.parse("https://www.youtube.com/watch?v=valid_id", result_format=ResultFormat.TXT)

@pytest.fixture
def mock_gitingest_error():
    """Mock gitingest to simulate repository not found error."""
    with patch('src.content_parser.text_parser.ingest') as MockIngest:
        MockIngest.side_effect = Exception("Repository not found")
        yield MockIngest

def test_text_parser_github_error(mock_gitingest_error):
    """Test that parser handles GitHub repository not found error."""
    parser = TextParser()
    with pytest.raises(Exception, match="Repository not found"):
        parser.parse("https://github.com/invalid/repo", result_format=ResultFormat.TXT)



from .base_parser import BaseParser
from markitdown import MarkItDown
from io import BytesIO
from ..utils import ResultFormat
import tempfile

class DataParser(BaseParser):
    def __init__(self):
        self.converter = MarkItDown()

    @property
    def supported_mimetypes(self) -> list:
        return [
            "text/csv",  # CSV
            "application/json",  # JSON
            "application/xml",  # XML
            "text/xml",  # XML
        ]

    def parse(self, file_content, result_format: ResultFormat) -> str:
        # Handle both bytes and file-like objects
        if hasattr(file_content, 'read'):
            content = file_content.read()
        else:
            content = file_content

        # Use file extension directly since these are text-based formats
        if b'<?xml' in content[:10] or b'<rss' in content[:10]:
            extension = '.xml'
        elif b'{' in content[:10]:
            extension = '.json'
        else:
            extension = '.csv'

        # MarkItDown expects a file path, so we need to write to a temp file
        with tempfile.NamedTemporaryFile(suffix=extension) as temp_file:
            temp_file.write(content)
            temp_file.flush()
            result = self.converter.convert(temp_file.name)
            text = result.text_content
            if result_format == ResultFormat.MD:
                if extension == '.json':
                    text = f"```json\n{text}\n```"
                elif extension == '.xml':
                    text = f"```xml\n{text}\n```"
            return text



import unittest
from src.parsethisio import get_supported_extensions

class TestSupportedExtensions(unittest.TestCase):
    def test_supported_extensions(self):
        # Get all supported extensions
        exts = get_supported_extensions()
        
        # Verify expected extensions are present
        self.assertIn('.pdf', exts)  # PDFParser
        self.assertIn('.jpg', exts)  # ImageParser
        self.assertIn('.png', exts)  # ImageParser
        self.assertIn('.mp3', exts)  # AudioParser
        self.assertIn('.txt', exts)  # TextParser
        
        # Verify extensions are properly formatted
        for ext in exts:
            self.assertTrue(ext.startswith('.'), f"Extension {ext} should start with a dot")



from .base_parser import BaseParser
from typing import BinaryIO
from openai import OpenAI
import base64
from ..utils import ResultFormat
from loguru import logger
from ..exceptions import ParsingFailed
'''
Approach: Send image to openAI, get markdown/text back - return it
'''
class ImageParser(BaseParser):
    @property
    def supported_mimetypes(self) -> list:
        return [
            "image/jpeg",
            "image/png"
        ]

    def parse(self, file_obj: BinaryIO, result_format: ResultFormat) -> str:
        client = OpenAI()

        #image file_content as base64 string:
        base64_image = base64.b64encode(file_obj).decode('utf-8')

        prompt_result_hint = ""
        if result_format == ResultFormat.MD:
            prompt_result_hint = "Answer in markdown format."
        elif result_format == ResultFormat.TXT:
            prompt_result_hint = "Answer in text format."

        response = client.chat.completions.create(
            model="gpt-4o-mini",
            messages=[
                {
                "role": "user",
                "content": [
                    {
                    "type": "text",
                    "text": "What is in this image? When the image contains a Diagram, give me the diagram as mermaid code. Give me the mermaid code only then, without any description. " + prompt_result_hint,
                    },
                    {
                    "type": "image_url",
                    "image_url": {
                        "url":  f"data:image/jpeg;base64,{base64_image}"
                    },
                    },
                ],
                }
            ],
        )

        if len(response.choices) == 0:
            logger.error("No response from openai")

            raise ParsingFailed("Image parsing failed")
        else:
            logger.debug(response.choices[0])

            return response.choices[0].message.content
        


from .base_parser import BaseParser
from typing import BinaryIO
from PyPDF2 import PdfReader
import io
from openai import OpenAI
import base64
from ..utils import ResultFormat

class AudioParser(BaseParser):
    @property
    def supported_mimetypes(self) -> list:
        return [
            "audio/ogg",
            "audio/mpeg",
            "audio/mp3",
            "audio/mp4",
            "audio/mpga",
            "audio/m4a",
            "audio/wav",
            "audio/webm",
        ]

    def parse(self, file_content: BinaryIO, result_format: ResultFormat) -> str:
        client = OpenAI()

        if isinstance(file_content, str):
            file_content = file_content.encode('utf-8')
        tmp_file_content = io.BytesIO(file_content)

        #TOOD: remove hardcoded "audio/mpeg" definition
        transcription = client.audio.transcriptions.create(
            model="whisper-1", 
            file=("test.mp3", tmp_file_content, "audio/mpeg")
        )

        #for this parser, text and markdown result is equal
        return transcription.text


import mimetypes
import filetype

#available parsers
from .content_parser.audio_parser import AudioParser
from .content_parser.image_parser import ImageParser
from .content_parser.pdf_parser import PDFParser
from .content_parser.text_parser import TextParser
from .content_parser.office_parser import OfficeParser
from .content_parser.data_parser import DataParser
from .content_parser.archive_parser import ArchiveParser

from .exceptions import ParserNotFoundError, UnsupportedMimeTypeError
from .utils import ResultFormat
PARSERS = [
    AudioParser,
    ImageParser,
    PDFParser,
    TextParser,
    OfficeParser,
    DataParser,
    ArchiveParser,
]

def parse(binary_file_data: bytes, file_name: str = None, mime_type: str = None, result_format: ResultFormat = ResultFormat.TXT):
    """
    Parse the given input source.

    This function takes an input source and performs parsing operations.
    When you dont provide file_name and mime_type, it will try to identify the file type automatically.
    When you encounter problems with automatic identification, you can provide file_name and mime_type manually.

    Args:
        binary_file_data (str): The source to be parsed.
        file_name (str): The filename will help to identify file type automatically.
        mime_type (str): The mime type will help to identify file type automatically

    Returns:
        str || None: The parsed content.

    Raises:
        ParserNotFoundError: If no parser is found for the given input.
        UnsupportedMimeTypeError: If the MIME type is unsupported.
        ParseThisError: For parsethis specific errors from sublogics
    """

    parser = get_parser(binary_file_data, file_name, mime_type)

    return parser.parse(binary_file_data, result_format)

def get_parser(binary_file_data: bytes, file_name: str = None, mime_type: str = None):
    """
    Determines the appropriate parser for the given input source.

    This function attempts to identify the appropriate parser based on the provided MIME type,
    file name, or the content of the binary file data. It prioritizes the MIME type if provided,
    followed by the file name, and finally attempts to guess the type based on the file content.

    Args:
        binary_file_data (bytes): The binary data of the file to be parsed.
        file_name (str, optional): The name of the file, used to help identify the file type.
        mime_type (str, optional): The MIME type of the file, used to identify the file type.

    Returns:
        Parser: The parser object that can handle the given input.

    Raises:
        ParserNotFoundError: If no parser is found for the given input.
        UnsupportedMimeTypeError: If the MIME type is unsupported.
    """
    #mime_type is most specific, lets use it if it's provided
    if mime_type is not None:
        return get_parser_by_mime_type(mime_type)
    
    #file_name is less specific, lets use it if it's provided
    if file_name is not None:
        mime_type = mimetypes.guess_type(file_name)[0]
        return get_parser_by_mime_type(mime_type)

    #if we dont have any information, lets try to guess the type based on file content
    kind = filetype.guess(binary_file_data)
    if kind is not None:
        return get_parser_by_mime_type(kind.mime)

    raise ParserNotFoundError("No parser found for the given input")

def get_parser_by_mime_type(mime_type: str):
    """
    Reads all Parser in content_parser folder, loads available mime_types and 
    returns the matching parser for the given mime_type.

    Args:
        mime_type (str): The MIME type to find a parser for.

    Returns:
        Parser: The parser object that can handle the given MIME type.

    Raises:
        UnsupportedMimeTypeError: If no parser supports the given MIME type.
    """

    for parser in PARSERS:
        parserObj = parser()
        if mime_type in parserObj.supported_mimetypes:
            return parserObj

    raise UnsupportedMimeTypeError("Unsupported MIME type: {mime_type}")

def get_supported_extensions() -> list[str]:
    """
    Get a list of all file extensions supported by any parser.
    
    Returns:
        list[str]: List of supported file extensions including the dot (e.g. ['.pdf', '.jpg'])
    """
    extensions = set()
    for parser in PARSERS:
        parser_obj = parser()
        for mime_type in parser_obj.supported_mimetypes:
            # Get all extensions for this mime type
            if mime_type:  # Skip None values
                exts = mimetypes.guess_all_extensions(mime_type)
                if exts:  # Skip empty lists
                    extensions.update(exts)
    return sorted(list(extensions))

# Pre-compute the list of extensions
supported_extensions = get_supported_extensions()




from abc import ABC, abstractmethod
from ..utils import ResultFormat
class BaseParser(ABC):
    @abstractmethod
    def parse(self, source, result_format: ResultFormat = ResultFormat.TXT) -> str:
        """Parse text from the given source."""
        pass

    @property
    @abstractmethod
    def supported_mimetypes(self) -> list:
        """List of supported MIME types."""
        pass



import os
from src.content_parser.office_parser import OfficeParser
from src.utils import ResultFormat
import filetype

def test_docx_parser():
    file_path = os.path.join(os.path.dirname(__file__), 'fixtures', 'test_data_sample.docx')
    
    with open(file_path, "rb") as f:
        content = f.read()
        mime = filetype.guess(content).mime
        assert mime == "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
        
        parser = OfficeParser()
        # Test TXT format
        text = parser.parse(content, result_format=ResultFormat.TXT)
        assert isinstance(text, str)
        assert len(text) > 0
        assert "AutoGen: Enabling Next-Gen LLM Applications" in text

        # Test MD format
        md_text = parser.parse(content, result_format=ResultFormat.MD)
        assert isinstance(md_text, str)
        assert len(md_text) > 0
        assert "# AutoGen: Enabling Next-Gen LLM Applications" in md_text

def test_pptx_parser():
    file_path = os.path.join(os.path.dirname(__file__), 'fixtures', 'test_data_presentation.pptx')
    
    with open(file_path, "rb") as f:
        content = f.read()
        mime = filetype.guess(content).mime
        assert mime == "application/vnd.openxmlformats-officedocument.presentationml.presentation"
        
        parser = OfficeParser()
        text = parser.parse(content, result_format=ResultFormat.TXT)
        assert isinstance(text, str)
        assert len(text) > 0
        assert "AutoGen: Enabling Next-Gen LLM Applications" in text

def test_xlsx_parser():
    file_path = os.path.join(os.path.dirname(__file__), 'fixtures', 'test_data_spreadsheet.xlsx')
    
    with open(file_path, "rb") as f:
        content = f.read()
        mime = filetype.guess(content).mime
        assert mime == "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
        
        parser = OfficeParser()
        text = parser.parse(content, result_format=ResultFormat.TXT)
        assert isinstance(text, str)
        assert len(text) > 0
        assert "Sheet1" in text

def test_xls_parser():
    file_path = os.path.join(os.path.dirname(__file__), 'fixtures', 'test_data_spreadsheet.xls')
    
    with open(file_path, "rb") as f:
        content = f.read()
        mime = filetype.guess(content).mime
        assert mime == "application/vnd.ms-excel"
        
        parser = OfficeParser()
        text = parser.parse(content, result_format=ResultFormat.TXT)
        assert isinstance(text, str)
        assert len(text) > 0
        assert "Sheet1" in text



import pytest
from unittest.mock import patch, MagicMock
from src.content_parser.image_parser import ImageParser
from src.parsethisio import ResultFormat
# src/content_parser/test_image_parser.py
import os

image_test_file_path = 'tests/fixtures/test_data_diagram.png'
image_test_file_as_text = '''
```mermaid
graph LR
    xychart-beta
    title "Sales Revenue"
    x-axis [jan, feb, mar, apr, may, jun, jul, aug, sep, oct, nov, dec]
    y-axis "Revenue (in $)" 4000 --> 11000
    bar [5000, 6000, 7500, 8200, 9500, 10500, 11000, 10200, 9200, 8500, 7000, 6000]
    line [5000, 6000, 7500, 8200, 9500, 10500, 11000, 10200, 9200, 8500, 7000, 6000]
```
'''


@pytest.fixture
def mock_openai():
    with patch('src.content_parser.image_parser.OpenAI') as MockOpenAI:
        mock_client = MockOpenAI.return_value
        mock_client.chat.completions.create.return_value = MagicMock(
            choices=[MagicMock(message=MagicMock(content=image_test_file_as_text))]
        )
        yield mock_client

def test_image_parser(mock_openai):
    parser = ImageParser()
    file_path = os.path.join(os.path.dirname(__file__), 'fixtures', 'test_data_diagram.png')
    
    with open(file_path, "rb") as f:
        result = parser.parse(f.read(), result_format=ResultFormat.TXT)

    assert result == image_test_file_as_text
    mock_openai.chat.completions.create.assert_called_once()


import re
import requests
from bs4 import BeautifulSoup, Tag
from youtube_transcript_api import YouTubeTranscriptApi
from loguru import logger
from ...exceptions import RegexResultError, RemoteRequestError, NotFoundError

class YouTubeTranscriptHelper:
    """Helper class to manage YouTube transcript extraction and metadata retrieval."""

    @staticmethod
    def extract_video_id(url: str) -> str:
        """
        Extracts the YouTube video ID from various URL formats.

        Args:
            url: The YouTube video URL

        Returns:
            str: The YouTube video ID

        Raises:
            RegexResultError: When regex wasnt able to find videoId
        """
        youtube_regex = (
            r"(?:https?://)?(?:www\.)?"
            r"(?:youtu\.be/|youtube\.com(?:/embed/|/v/|/watch\?v=|/watch\?.+&v=|/shorts/))"
            r"([\w-]{11})"
        )
        match = re.search(youtube_regex, url)
        if not match:
            raise RegexResultError(f"Failed to extract video ID from {url}")

        return match.group(1)

    @staticmethod
    def get_video_title(video_id: str) -> str:
        """
        Fetches the title of the YouTube video using the video ID.

        Args:
            video_id: The YouTube video ID

        Returns:
            str: The video title

        Raises:
            RemoteRequestError: When we were not able to fetch the youtube url
            RegexResultError: When we cant parse the title from the html
        """
        try:
            url = f"https://www.youtube.com/watch?v={video_id}"
            response = requests.get(url)
            html = response.text

            soup = BeautifulSoup(html, "html.parser")
            meta = soup.find("meta", property="og:title")
            if not meta or not isinstance(meta, Tag) or not meta.get("content"):
                raise RegexResultError(f"Failed to extract video title from {url}")
            
            content = meta.get("content")
            if not isinstance(content, str):
                raise RegexResultError(f"Invalid title content type from {url}")
            
            return content
        except Exception as e:
            raise RemoteRequestError(f"Failed to fetch video title for {video_id}: {e}")

    #TODO: update preferred langs?
    @staticmethod
    def get_best_transcript(video_id: str, preferred_langs: list[str] = ["en", "es", "pt"]) -> list[dict[str, str | float]]:
        """
        Retrieves the best transcript for a YouTube video.

        Args:
            video_id: The YouTube video ID
            preferred_langs: Preferred languages for the transcript

        Returns:
            list[dict[str, str | float]]: The transcript entries with text and timing info

        Raises:
            NotFoundError: When no suitable transcript is found
            RemoteRequestError: When we were not able to fetch the youtube url
        """
        try:
            transcript_list = YouTubeTranscriptApi.list_transcripts(video_id)

            # Prioritize manual transcripts
            for lang in preferred_langs:
                for transcript in transcript_list:
                    if not transcript.is_generated and transcript.language_code == lang:
                        return transcript.fetch()

            # Fallback to auto-generated transcripts
            for lang in preferred_langs:
                for transcript in transcript_list:
                    if transcript.is_generated and transcript.language_code == lang:
                        return transcript.fetch()

            # Fallback to translated transcripts
            for transcript in transcript_list:
                if transcript.is_translatable:
                    translation = transcript.translate(preferred_langs[0])
                    return translation.fetch()

            raise NotFoundError("No suitable transcript found")
        except Exception as e:
            raise RemoteRequestError(f"Failed to get transcript for video {video_id}: {e}")

    @staticmethod
    def clean_transcript(transcript: list[dict[str, str | float]], remove_phrases: list[str] | None = None) -> str:
        """
        Cleans the transcript by removing unwanted phrases or tags.

        Args:
            transcript: The raw transcript entries with text and timing info
            remove_phrases: A list of phrases to remove (case-insensitive)

        Returns:
            str: The cleaned transcript as a single string
        """
        if not remove_phrases:
            remove_phrases = ["[music]", "[applause]", "[laughter]"]

        cleaned_texts: list[str] = []
        for entry in transcript:
            text = entry.get("text")
            if not isinstance(text, str):
                continue
            
            cleaned_text = text
            for phrase in remove_phrases:
                cleaned_text = cleaned_text.replace(phrase, "").strip()
            
            if cleaned_text:
                cleaned_texts.append(cleaned_text)

        return " ".join(cleaned_texts)
